/*
 * Sonar, entreprise quality control tool.
 * Copyright (C) 2007-2008 Hortis-GRC SA
 * mailto:be_agile HAT hortis DOT ch
 *
 * Sonar is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * Sonar is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with Sonar; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02
 */
package org.sonar.plugins.clover;

import ch.hortis.sonar.service.MeasureKey;
import org.hibernate.Hibernate;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.sonar.commons.measures.MeasureParameter;
import org.sonar.plugins.api.maven.JavaMeasuresRecorder;
import org.sonar.plugins.api.maven.MavenCollectorUtils;
import org.sonar.plugins.api.maven.MavenUtils;
import org.sonar.plugins.api.maven.xml.XmlParserException;
import org.sonar.plugins.api.maven.xml.XmlReportParser;
import org.sonar.plugins.api.metrics.CoreMetrics;
import org.sonar.plugins.core.unittests.CodeCoverageUtils;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.apache.commons.lang.StringUtils;

import javax.xml.xpath.XPathConstants;
import java.io.File;
import java.sql.Blob;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.List;

public class XmlCollector {

  private static final Logger LOG = LoggerFactory.getLogger(XmlCollector.class);
  private JavaMeasuresRecorder measuresRecorder;

  public XmlCollector(JavaMeasuresRecorder measuresRecorder) {
    this.measuresRecorder = measuresRecorder;
  }

  protected void collect(File xmlReportFile) {
    try {
      if (xmlReportFile != null && xmlReportFile.exists()) {
        XmlReportParser parser = new XmlReportParser();
        parser.parse(xmlReportFile);
        collectProjectMeasures(parser);
        collectPackageMeasures(parser);
        collectFileMeasures(parser);
      }
    } catch (ParseException e) {
      throw new XmlParserException(e);
    }
  }

  private void collectProjectMeasures(XmlReportParser parser) throws ParseException {
    Double result = null;
    try {
      Element projectEl = parser.getChildElement(parser.getRoot(), "project");
      Element metricsEl = parser.getChildElement(projectEl, "metrics");
      result = getCodeCoverageFromMetricsNode(metricsEl);
      measuresRecorder.createProjectMeasure(new MeasureKey(CoreMetrics.CODE_COVERAGE), result);
    } catch (IllegalArgumentException e) {
      LOG.warn("result is invalid : ", result);
    }
  }

  private void collectPackageMeasures(XmlReportParser parser) throws ParseException {
    Double result = null;
    try {
      NodeList packages = parser.executeXPathNodeList("/coverage/project/package");
      for (int i = 0; i < packages.getLength(); i++) {
        Element pkElt = (Element) packages.item(i);
        Element metricsEl = parser.getChildElement(pkElt, "metrics");
        result = getCodeCoverageFromMetricsNode(metricsEl);
        measuresRecorder.createPackageMeasure(pkElt.getAttribute("name"), new MeasureKey(CoreMetrics.CODE_COVERAGE), result);
      }
    } catch (IllegalArgumentException e) {
      LOG.warn("result is invalid : ", result);
    }
  }

  private void collectFileMeasures(XmlReportParser parser) throws ParseException {
    Double result = null;
    try {
      NodeList packages = parser.executeXPathNodeList("/coverage/project/package");
      for (int i = 0; i < packages.getLength(); i++) {
        Element pkElt = (Element) packages.item(i);
        String namespace = pkElt.getAttribute("name");
        NodeList files = parser.executeXPathNodeList(pkElt, "file");
        for (int j = 0; j < files.getLength(); j++) {
          Element fileElt = (Element) files.item(j);
          String filename = extractClassName(fileElt);
          if (filename != null) {
            Element metricsElt = (Element) parser.executeXPath(fileElt, XPathConstants.NODE, "metrics");
            if (canBeIncludedInFileMetrics(metricsElt)) {
              result = getCodeCoverageFromMetricsNode(metricsElt);
              measuresRecorder.createClassMeasure(namespace, filename, false, new MeasureKey(CoreMetrics.CODE_COVERAGE), result,
                  getParameters(parser, fileElt));
            }
          }
        }
      }
    } catch (IllegalArgumentException e) {
      LOG.warn("result is invalid : ", result);
    }
  }

  private MeasureParameter[] getParameters(XmlReportParser parser, Element fileElt) throws ParseException {
    List<MeasureParameter> parameters = new ArrayList<MeasureParameter>();
    List<Element> lines = parser.getChildElements(fileElt, "line");
    StringBuilder levelsByLine = new StringBuilder();
    for (Element line : lines) {
      double hits;
      if ("".equals(line.getAttribute("count"))) {
        hits = MavenCollectorUtils.parseNumber(line.getAttribute("truecount")) + MavenCollectorUtils.parseNumber(line.getAttribute("falsecount"));
      } else {
        hits = MavenCollectorUtils.parseNumber(line.getAttribute("count"));
      }
      CodeCoverageUtils.appendParameter(levelsByLine, line.getAttribute("num"), Integer.toString((int) hits));
    }
    Blob lobValue = Hibernate.createBlob(levelsByLine.toString().getBytes());
    parameters.add(new MeasureParameter("lines", lobValue, null));
    return parameters.toArray(new MeasureParameter[0]);
  }

  private boolean canBeIncludedInFileMetrics(Element metricsNode) throws ParseException {
    return MavenCollectorUtils.parseNumber(metricsNode.getAttribute("elements")) > 0;
  }

  private Double getCodeCoverageFromMetricsNode(Element metricsNode) throws ParseException {
    double coveredElements = MavenCollectorUtils.parseNumber(metricsNode.getAttribute("coveredelements"));
    double nb = MavenCollectorUtils.parseNumber(metricsNode.getAttribute("elements"));
    return 100.0 * (coveredElements / nb);
  }

  public static String extractClassName(Element fileElt) {
    String filename = fileElt.getAttribute("name");
    if (filename != null) {
      filename = StringUtils.replaceChars(filename, '\\', '/');
      filename = StringUtils.substringBeforeLast(filename, ".java");
      if (filename.indexOf("/")>=0) {
        filename = StringUtils.substringAfterLast(filename, "/");
      }
    }
    return filename;
  }

}
